import threading
import unittest
import uuid

from pykka import ActorDeadError, ActorRegistry, ThreadingActor


class AnActor(object):

    def __init__(self, **events):
        super(AnActor, self).__init__()
        self.on_start_was_called = events['on_start_was_called']
        self.on_stop_was_called = events['on_stop_was_called']
        self.on_failure_was_called = events['on_failure_was_called']
        self.actor_was_registered_before_on_start_was_called = events[
            'actor_was_registered_before_on_start_was_called']
        self.greetings_was_received = events['greetings_was_received']

    def on_start(self):
        self.on_start_was_called.set()
        if ActorRegistry.get_by_urn(self.actor_urn) is not None:
            self.actor_was_registered_before_on_start_was_called.set()

    def on_stop(self):
        self.on_stop_was_called.set()

    def on_failure(self, *args):
        self.on_failure_was_called.set()

    def on_receive(self, message):
        if message.get('command') == 'raise exception':
            raise Exception('foo')
        elif message.get('command') == 'raise base exception':
            raise BaseException()
        elif message.get('command') == 'stop twice':
            self.stop()
            self.stop()
        elif message.get('command') == 'message self then stop':
            self.actor_ref.tell({'command': 'greetings'})
            self.stop()
        elif message.get('command') == 'greetings':
            self.greetings_was_received.set()
        elif message.get('command') == 'callback':
            message['callback']()
        else:
            super(AnActor, self).on_receive(message)


class EarlyStoppingActor(object):

    def __init__(self, on_stop_was_called):
        super(EarlyStoppingActor, self).__init__()
        self.on_stop_was_called = on_stop_was_called

    def on_start(self):
        self.stop()

    def on_stop(self):
        self.on_stop_was_called.set()


class EarlyFailingActor(object):

    def __init__(self, on_start_was_called):
        super(EarlyFailingActor, self).__init__()
        self.on_start_was_called = on_start_was_called

    def on_start(self):
        try:
            raise RuntimeError('on_start failure')
        finally:
            self.on_start_was_called.set()


class LateFailingActor(object):

    def __init__(self, on_stop_was_called):
        super(LateFailingActor, self).__init__()
        self.on_stop_was_called = on_stop_was_called

    def on_start(self):
        self.stop()

    def on_stop(self):
        try:
            raise RuntimeError('on_stop failure')
        finally:
            self.on_stop_was_called.set()


class FailingOnFailureActor(object):

    def __init__(self, on_failure_was_called):
        super(FailingOnFailureActor, self).__init__()
        self.on_failure_was_called = on_failure_was_called

    def on_receive(self, message):
        if message.get('command') == 'raise exception':
            raise Exception('on_receive failure')
        else:
            super(FailingOnFailureActor, self).on_receive(message)

    def on_failure(self, *args):
        try:
            raise RuntimeError('on_failure failure')
        finally:
            self.on_failure_was_called.set()


class ActorTest(object):

    def setUp(self):
        self.on_start_was_called = self.event_class()
        self.on_stop_was_called = self.event_class()
        self.on_failure_was_called = self.event_class()
        self.actor_was_registered_before_on_start_was_called = (
            self.event_class())
        self.greetings_was_received = self.event_class()

        self.actor_ref = self.AnActor.start(
            on_start_was_called=self.on_start_was_called,
            on_stop_was_called=self.on_stop_was_called,
            on_failure_was_called=self.on_failure_was_called,
            actor_was_registered_before_on_start_was_called=(
                self.actor_was_registered_before_on_start_was_called),
            greetings_was_received=self.greetings_was_received)
        self.actor_proxy = self.actor_ref.proxy()

    def tearDown(self):
        ActorRegistry.stop_all()

    def test_messages_left_in_queue_after_actor_stops_receive_an_error(self):
        event = self.event_class()
        self.actor_ref.tell({'command': 'callback', 'callback': event.wait})
        self.actor_ref.stop(block=False)
        response = self.actor_ref.ask({'command': 'irrelevant'}, block=False)
        event.set()

        self.assertRaises(ActorDeadError, response.get, timeout=0.5)

    def test_stop_requests_left_in_queue_after_actor_stops_are_handled(self):
        event = self.event_class()
        self.actor_ref.tell({'command': 'callback', 'callback': event.wait})
        self.actor_ref.stop(block=False)
        response = self.actor_ref.ask({'command': 'pykka_stop'}, block=False)
        event.set()

        response.get(timeout=0.5)

    def test_actor_has_an_uuid4_based_urn(self):
        self.assertEqual(4, uuid.UUID(self.actor_ref.actor_urn).version)

    def test_actor_has_unique_uuid(self):
        event = self.event_class()
        actors = [
            self.AnActor.start(
                on_start_was_called=event,
                on_stop_was_called=event,
                on_failure_was_called=event,
                actor_was_registered_before_on_start_was_called=event,
                greetings_was_received=event)
            for _ in range(3)]

        self.assertNotEqual(actors[0].actor_urn, actors[1].actor_urn)
        self.assertNotEqual(actors[1].actor_urn, actors[2].actor_urn)
        self.assertNotEqual(actors[2].actor_urn, actors[0].actor_urn)

    def test_str_on_raw_actor_contains_actor_class_name(self):
        event = self.event_class()
        unstarted_actor = self.AnActor(
            on_start_was_called=event,
            on_stop_was_called=event,
            on_failure_was_called=event,
            actor_was_registered_before_on_start_was_called=event,
            greetings_was_received=event)
        self.assert_('AnActor' in str(unstarted_actor))

    def test_str_on_raw_actor_contains_actor_urn(self):
        event = self.event_class()
        unstarted_actor = self.AnActor(
            on_start_was_called=event,
            on_stop_was_called=event,
            on_failure_was_called=event,
            actor_was_registered_before_on_start_was_called=event,
            greetings_was_received=event)
        self.assert_(unstarted_actor.actor_urn in str(unstarted_actor))

    def test_init_can_be_called_with_arbitrary_arguments(self):
        self.SuperInitActor(1, 2, 3, foo='bar')

    def test_on_start_is_called_before_first_message_is_processed(self):
        self.on_start_was_called.wait(5)
        self.assertTrue(self.on_start_was_called.is_set())

    def test_on_start_is_called_after_the_actor_is_registered(self):
        # NOTE: If the actor is registered after the actor is started, this
        # test may still occasionally pass, as it is dependant on the exact
        # timing of events. When the actor is first registered and then
        # started, this test should always pass.
        self.on_start_was_called.wait(5)
        self.assertTrue(self.on_start_was_called.is_set())
        self.actor_was_registered_before_on_start_was_called.wait(0.1)
        self.assertTrue(
            self.actor_was_registered_before_on_start_was_called.is_set())

    def test_on_start_can_stop_actor_before_receive_loop_is_started(self):
        # NOTE: This test will pass even if the actor is allowed to start the
        # receive loop, but it will cause the test suite to hang, as the actor
        # thread is blocking on receiving messages to the actor inbox forever.
        # If one made this test specifically for ThreadingActor, one could add
        # an assertFalse(actor_thread.is_alive()), which would cause the test
        # to fail properly.
        stop_event = self.event_class()
        another_actor = self.EarlyStoppingActor.start(stop_event)
        stop_event.wait(5)
        self.assertTrue(stop_event.is_set())
        self.assertFalse(another_actor.is_alive())

    def test_on_start_failure_causes_actor_to_stop(self):
        # Actor should not be alive if on_start fails.
        start_event = self.event_class()
        actor_ref = self.EarlyFailingActor.start(start_event)
        start_event.wait(5)
        actor_ref.actor_stopped.wait(5)
        self.assertFalse(actor_ref.is_alive())

    def test_on_stop_is_called_when_actor_is_stopped(self):
        self.assertFalse(self.on_stop_was_called.is_set())
        self.actor_ref.stop()
        self.on_stop_was_called.wait(5)
        self.assertTrue(self.on_stop_was_called.is_set())

    def test_on_stop_failure_causes_actor_to_stop(self):
        stop_event = self.event_class()
        actor = self.LateFailingActor.start(stop_event)
        stop_event.wait(5)
        self.assertFalse(actor.is_alive())

    def test_on_failure_is_called_when_exception_cannot_be_returned(self):
        self.assertFalse(self.on_failure_was_called.is_set())
        self.actor_ref.tell({'command': 'raise exception'})
        self.on_failure_was_called.wait(5)
        self.assertTrue(self.on_failure_was_called.is_set())
        self.assertFalse(self.on_stop_was_called.is_set())

    def test_on_failure_failure_causes_actor_to_stop(self):
        failure_event = self.event_class()
        actor = self.FailingOnFailureActor.start(failure_event)
        actor.tell({'command': 'raise exception'})
        failure_event.wait(5)
        self.assertFalse(actor.is_alive())

    def test_actor_is_stopped_when_unhandled_exceptions_are_raised(self):
        self.assertFalse(self.on_failure_was_called.is_set())
        self.actor_ref.tell({'command': 'raise exception'})
        self.on_failure_was_called.wait(5)
        self.assertTrue(self.on_failure_was_called.is_set())
        self.assertEqual(0, len(ActorRegistry.get_all()))

    def test_all_actors_are_stopped_on_base_exception(self):
        start_event = self.event_class()
        stop_event = self.event_class()
        fail_event = self.event_class()
        registered_event = self.event_class()
        greetings_event = self.event_class()
        self.AnActor.start(
            on_start_was_called=start_event,
            on_stop_was_called=stop_event,
            on_failure_was_called=fail_event,
            actor_was_registered_before_on_start_was_called=registered_event,
            greetings_was_received=greetings_event)

        self.assertEqual(2, len(ActorRegistry.get_all()))
        self.assertFalse(self.on_stop_was_called.is_set())
        self.actor_ref.tell({'command': 'raise base exception'})
        self.on_stop_was_called.wait(5)
        self.assertTrue(self.on_stop_was_called.is_set())
        self.assert_(1 >= len(ActorRegistry.get_all()))
        stop_event.wait(5)
        self.assertTrue(stop_event.is_set())
        self.assertEqual(0, len(ActorRegistry.get_all()))

    def test_actor_can_call_stop_on_self_multiple_times(self):
        self.actor_ref.ask({'command': 'stop twice'})

    def test_actor_processes_all_messages_before_stop_on_self_stops_it(self):
        self.actor_ref.ask({'command': 'message self then stop'})

        self.greetings_was_received.wait(5)
        self.assertTrue(self.greetings_was_received.is_set())

        self.on_stop_was_called.wait(5)

        self.assertEqual(0, len(ActorRegistry.get_all()))


def ConcreteActorTest(actor_class, event_class):
    class C(ActorTest, unittest.TestCase):

        class AnActor(AnActor, actor_class):
            pass

        class EarlyStoppingActor(EarlyStoppingActor, actor_class):
            pass

        class EarlyFailingActor(EarlyFailingActor, actor_class):
            pass

        class LateFailingActor(LateFailingActor, actor_class):
            pass

        class FailingOnFailureActor(FailingOnFailureActor, actor_class):
            pass

        class SuperInitActor(actor_class):
            pass

    C.__name__ = '%sTest' % actor_class.__name__
    C.event_class = event_class
    return C


class ThreadingActorTest(ConcreteActorTest(ThreadingActor, threading.Event)):

    class DaemonActor(ThreadingActor):
        use_daemon_thread = True

    def test_actor_thread_is_named_after_pykka_actor_class(self):
        alive_threads = threading.enumerate()
        alive_thread_names = [t.name for t in alive_threads]
        named_correctly = [
            name.startswith(AnActor.__name__) for name in alive_thread_names]
        self.assert_(any(named_correctly))

    def test_actor_thread_is_not_daemonic_by_default(self):
        alive_threads = threading.enumerate()
        actor_threads = [
            t for t in alive_threads if t.name.startswith('AnActor')]
        self.assertEqual(1, len(actor_threads))
        self.assertFalse(actor_threads[0].daemon)

    def test_actor_thread_is_daemonic_if_use_daemon_thread_flag_is_set(self):
        actor_ref = self.DaemonActor.start()
        alive_threads = threading.enumerate()
        actor_threads = [
            t for t in alive_threads if t.name.startswith('DaemonActor')]
        self.assertEqual(1, len(actor_threads))
        self.assertTrue(actor_threads[0].daemon)
        actor_ref.stop()


try:
    import gevent.event
    from pykka.gevent import GeventActor
    GeventActorTest = ConcreteActorTest(GeventActor, gevent.event.Event)
except ImportError:
    pass

try:
    import eventlet  # noqa
    from pykka.eventlet import EventletActor, EventletEvent
    EventletActorTest = ConcreteActorTest(EventletActor, EventletEvent)
except ImportError:
    pass



# encoding: utf-8

"""Pykka documentation build configuration file"""

from __future__ import unicode_literals

import os
import re
import sys


# -- Workarounds to have autodoc generate API docs ----------------------------

sys.path.insert(0, os.path.abspath('..'))


class Mock(object):

    def __init__(self, *args, **kwargs):
        pass

    def __call__(self, *args, **kwargs):
        return Mock()

    @classmethod
    def __getattr__(self, name):
        if name in ('__file__', '__path__'):
            return '/dev/null'
        elif name[0] == name[0].upper():
            return type(name, (), {})
        else:
            return Mock()


MOCK_MODULES = [
    'gevent',
    'gevent.event',
    'gevent.queue',
    'eventlet',
    'eventlet.event',
    'eventlet.queue',
]
for mod_name in MOCK_MODULES:
    sys.modules[mod_name] = Mock()


# -- General configuration ----------------------------------------------------

needs_sphinx = '1.0'

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.extlinks',
    'sphinx.ext.intersphinx',
    'sphinx.ext.viewcode',
]

templates_path = ['_templates']
source_suffix = '.rst'
master_doc = 'index'

project = u'Pykka'
copyright = u'2010-2015, Stein Magnus Jodal'


def get_version():
    init_py = open('../pykka/__init__.py').read()
    metadata = dict(re.findall("__([a-z]+)__ = '([^']+)'", init_py))
    return metadata['version']


release = get_version()
version = '.'.join(release.split('.')[:2])

exclude_patterns = ['_build']

pygments_style = 'sphinx'

modindex_common_prefix = ['pykka.']


# -- Options for HTML output --------------------------------------------------

html_theme = 'default'
html_static_path = ['_static']

html_use_modindex = True
html_use_index = True
html_split_index = False
html_show_sourcelink = True

htmlhelp_basename = 'Pykka'


# -- Options for LaTeX output -------------------------------------------------

latex_documents = [
    (
        'index',
        'Pykka.tex',
        'Pykka Documentation',
        'Stein Magnus Jodal',
        'manual',
    ),
]


# -- Options for manual page output -------------------------------------------

man_pages = []


# -- Options for autodoc extension --------------------------------------------

autodoc_member_order = 'bysource'


# -- Options for extlink extension --------------------------------------------

extlinks = {
    'issue': ('https://github.com/jodal/pykka/issues/%s', '#'),
}


# -- Options for intersphinx extension ----------------------------------------

intersphinx_mapping = {
    'python': ('http://docs.python.org/2', None),
}



import sys

PY2 = sys.version_info[0] == 2
PY3 = sys.version_info[0] == 3

if PY2:
    import Queue as queue  # noqa

    string_types = basestring  # noqa

    def reraise(tp, value, tb=None):
        exec('raise tp, value, tb')

else:
    import queue  # noqa

    string_types = (str,)

    def reraise(tp, value, tb=None):
        if value is None:
            value = tp()
        if value.__traceback__ is not tb:
            raise value.with_traceback(tb)
        raise value



from __future__ import absolute_import

import sys

import eventlet
import eventlet.event
import eventlet.queue

from pykka import Timeout
from pykka.actor import Actor
from pykka.future import Future


__all__ = [
    'EventletActor',
    'EventletEvent',
    'EventletFuture',
]


class EventletEvent(eventlet.event.Event):

    """
    :class:`EventletEvent` adapts :class:`eventlet.event.Event` to
    :class:`threading.Event` interface.
    """

    def set(self):
        if self.ready():
            self.reset()
        self.send()

    def is_set(self):
        return self.ready()

    isSet = is_set

    def clear(self):
        if self.ready():
            self.reset()

    def wait(self, timeout):
        if timeout is not None:
            wait_timeout = eventlet.Timeout(timeout)

            try:
                with wait_timeout:
                    super(EventletEvent, self).wait()
            except eventlet.Timeout as t:
                if t is not wait_timeout:
                    raise
                return False
        else:
            self.event.wait()

        return True


class EventletFuture(Future):

    """
    :class:`EventletFuture` implements :class:`pykka.Future` for use with
    :class:`EventletActor`.
    """

    event = None

    def __init__(self):
        super(EventletFuture, self).__init__()
        self.event = eventlet.event.Event()

    def get(self, timeout=None):
        try:
            return super(EventletFuture, self).get(timeout=timeout)
        except NotImplementedError:
            pass

        if timeout is not None:
            wait_timeout = eventlet.Timeout(timeout)
            try:
                with wait_timeout:
                    return self.event.wait()
            except eventlet.Timeout as t:
                if t is not wait_timeout:
                    raise
                raise Timeout(t)
        else:
            return self.event.wait()

    def set(self, value=None):
        self.event.send(value)

    def set_exception(self, exc_info=None):
        if isinstance(exc_info, BaseException):
            exc_info = (exc_info,)
        self.event.send_exception(*(exc_info or sys.exc_info()))


class EventletActor(Actor):

    """
    :class:`EventletActor` implements :class:`pykka.Actor` using the `eventlet
    <http://eventlet.net/>`_ library.

    This implementation uses eventlet green threads.
    """

    @staticmethod
    def _create_actor_inbox():
        return eventlet.queue.Queue()

    @staticmethod
    def _create_future():
        return EventletFuture()

    def _start_actor_loop(self):
        eventlet.greenthread.spawn(self._actor_loop)



from pykka.actor import Actor, ActorRef
from pykka.exceptions import ActorDeadError, Timeout
from pykka.future import Future, get_all
from pykka.proxy import ActorProxy
from pykka.registry import ActorRegistry
from pykka.threading import ThreadingActor, ThreadingFuture


__all__ = [
    'Actor',
    'ActorDeadError',
    'ActorProxy',
    'ActorRef',
    'ActorRegistry',
    'Future',
    'ThreadingActor',
    'ThreadingFuture',
    'Timeout',
    'get_all',
]


#: Pykka's :pep:`396` and :pep:`440` compatible version number
__version__ = '1.2.1'


def _add_null_handler_for_logging():
    import logging
    try:
        NullHandler = logging.NullHandler  # Python 2.7 and upwards
    except AttributeError:
        class NullHandler(logging.Handler):

            def emit(self, record):
                pass
    logging.getLogger('pykka').addHandler(NullHandler())

_add_null_handler_for_logging()



import sys
import traceback
import unittest

from pykka import Future, ThreadingFuture, Timeout, get_all


class FutureBaseTest(unittest.TestCase):

    def setUp(self):
        self.future = Future()

    def test_future_get_is_not_implemented(self):
        self.assertRaises(NotImplementedError, self.future.get)

    def test_future_set_is_not_implemented(self):
        self.assertRaises(NotImplementedError, self.future.set, None)

    def test_future_set_exception_is_not_implemented(self):
        self.assertRaises(NotImplementedError, self.future.set_exception, None)


class FutureTest(object):

    def setUp(self):
        self.results = [self.future_class() for _ in range(3)]

    def test_set_multiple_times_fails(self):
        self.results[0].set(0)
        self.assertRaises(Exception, self.results[0].set, 0)

    def test_get_all_blocks_until_all_futures_are_available(self):
        self.results[0].set(0)
        self.results[1].set(1)
        self.results[2].set(2)
        result = get_all(self.results)
        self.assertEqual(result, [0, 1, 2])

    def test_get_all_raises_timeout_if_not_all_futures_are_available(self):
        self.results[0].set(0)
        self.results[1].set(1)

        self.assertRaises(Timeout, get_all, self.results, timeout=0)

    def test_get_all_can_be_called_multiple_times(self):
        self.results[0].set(0)
        self.results[1].set(1)
        self.results[2].set(2)
        result1 = get_all(self.results)
        result2 = get_all(self.results)
        self.assertEqual(result1, result2)

    def test_future_in_future_works(self):
        inner_future = self.future_class()
        inner_future.set('foo')
        outer_future = self.future_class()
        outer_future.set(inner_future)
        self.assertEqual(outer_future.get().get(), 'foo')

    def test_get_raises_exception_with_full_traceback(self):
        exc_class_get = None
        exc_class_set = None
        exc_instance_get = None
        exc_instance_set = None
        exc_traceback_get = None
        exc_traceback_set = None
        future = self.future_class()

        try:
            raise NameError('foo')
        except NameError:
            exc_class_set, exc_instance_set, exc_traceback_set = sys.exc_info()
            future.set_exception()

        # We could move to another thread at this point

        try:
            future.get()
        except NameError:
            exc_class_get, exc_instance_get, exc_traceback_get = sys.exc_info()

        self.assertEqual(exc_class_set, exc_class_get)
        self.assertEqual(exc_instance_set, exc_instance_get)

        exc_traceback_list_set = list(reversed(
            traceback.extract_tb(exc_traceback_set)))
        exc_traceback_list_get = list(reversed(
            traceback.extract_tb(exc_traceback_get)))

        # All frames from the first traceback should be included in the
        # traceback from the future.get() reraise
        self.assert_(len(exc_traceback_list_set) < len(exc_traceback_list_get))
        for i, frame in enumerate(exc_traceback_list_set):
            self.assertEquals(frame, exc_traceback_list_get[i])

    def test_filter_excludes_items_not_matching_predicate(self):
        future = self.results[0].filter(lambda x: x > 10)
        self.results[0].set([1, 3, 5, 7, 9, 11, 13, 15, 17, 19])

        self.assertEqual(future.get(timeout=0), [11, 13, 15, 17, 19])

    def test_filter_on_noniterable(self):
        future = self.results[0].filter(lambda x: x > 10)
        self.results[0].set(1)

        self.assertRaises(TypeError, future.get, timeout=0)

    def test_filter_preserves_the_timeout_kwarg(self):
        future = self.results[0].filter(lambda x: x > 10)

        self.assertRaises(Timeout, future.get, timeout=0)

    def test_join_combines_multiple_futures_into_one(self):
        future = self.results[0].join(self.results[1], self.results[2])
        self.results[0].set(0)
        self.results[1].set(1)
        self.results[2].set(2)

        self.assertEqual(future.get(timeout=0), [0, 1, 2])

    def test_join_preserves_timeout_kwarg(self):
        future = self.results[0].join(self.results[1], self.results[2])
        self.results[0].set(0)
        self.results[1].set(1)

        self.assertRaises(Timeout, future.get, timeout=0)

    def test_map_returns_future_which_passes_noniterable_through_func(self):
        future = self.results[0].map(lambda x: x + 10)
        self.results[0].set(30)

        self.assertEqual(future.get(timeout=0), 40)

    def test_map_returns_future_which_maps_iterable_through_func(self):
        future = self.results[0].map(lambda x: x + 10)
        self.results[0].set([10, 20, 30])

        self.assertEqual(future.get(timeout=0), [20, 30, 40])

    def test_map_preserves_timeout_kwarg(self):
        future = self.results[0].map(lambda x: x + 10)

        self.assertRaises(Timeout, future.get, timeout=0)

    def test_reduce_applies_function_cumulatively_from_the_left(self):
        future = self.results[0].reduce(lambda x, y: x + y)
        self.results[0].set([1, 2, 3, 4])

        self.assertEqual(future.get(timeout=0), 10)

    def test_reduce_accepts_an_initial_value(self):
        future = self.results[0].reduce(lambda x, y: x + y, 5)
        self.results[0].set([1, 2, 3, 4])

        self.assertEqual(future.get(timeout=0), 15)

    def test_reduce_on_noniterable(self):
        future = self.results[0].reduce(lambda x, y: x + y)
        self.results[0].set(1)

        self.assertRaises(TypeError, future.get, timeout=0)

    def test_reduce_preserves_the_timeout_kwarg(self):
        future = self.results[0].reduce(lambda x, y: x + y)

        self.assertRaises(Timeout, future.get, timeout=0)


class ThreadingFutureTest(FutureTest, unittest.TestCase):
    future_class = ThreadingFuture


try:
    from gevent.event import AsyncResult
    from pykka.gevent import GeventFuture

    class GeventFutureTest(FutureTest, unittest.TestCase):
        future_class = GeventFuture

        def test_can_wrap_existing_async_result(self):
            async_result = AsyncResult()
            future = GeventFuture(async_result)
            self.assertEquals(async_result, future.async_result)

        def test_get_raises_exception_with_full_traceback(self):
            # gevent prints the first half of the traceback instead of
            # passing it through to the other side of the AsyncResult
            pass
except ImportError:
    pass


try:
    from pykka.eventlet import EventletFuture

    class EventletFutureTest(FutureTest, unittest.TestCase):
        future_class = EventletFuture
except ImportError:
    pass



__all__ = [
    'ActorDeadError',
    'Timeout',
]


class ActorDeadError(Exception):

    """Exception raised when trying to use a dead or unavailable actor."""
    pass


class Timeout(Exception):

    """Exception raised at future timeout."""
    pass



import time

from pykka import ActorRegistry, ThreadingActor


def time_it(func):
    start = time.time()
    func()
    print('%s took %.3fs' % (func.func_name, time.time() - start))


class SomeObject(object):
    pykka_traversable = False
    cat = 'bar.cat'

    def func(self):
        pass


class AnActor(ThreadingActor):
    bar = SomeObject()
    bar.pykka_traversable = True

    foo = 'foo'

    def __init__(self):
        super(AnActor, self).__init__()
        self.cat = 'quox'

    def func(self):
        pass


def test_direct_plain_attribute_access():
    actor = AnActor.start().proxy()
    for _ in range(10000):
        actor.foo.get()


def test_direct_callable_attribute_access():
    actor = AnActor.start().proxy()
    for _ in range(10000):
        actor.func().get()


def test_traversible_plain_attribute_access():
    actor = AnActor.start().proxy()
    for _ in range(10000):
        actor.bar.cat.get()


def test_traversible_callable_attribute_access():
    actor = AnActor.start().proxy()
    for _ in range(10000):
        actor.bar.func().get()


if __name__ == '__main__':
    try:
        time_it(test_direct_plain_attribute_access)
        time_it(test_direct_callable_attribute_access)
        time_it(test_traversible_plain_attribute_access)
        time_it(test_traversible_callable_attribute_access)
    finally:
        ActorRegistry.stop_all()



from __future__ import absolute_import

import sys

import gevent
import gevent.event
import gevent.queue

from pykka import Timeout
from pykka.actor import Actor
from pykka.future import Future


__all__ = [
    'GeventActor',
    'GeventFuture',
]


class GeventFuture(Future):

    """
    :class:`GeventFuture` implements :class:`pykka.Future` for use with
    :class:`GeventActor`.

    It encapsulates a :class:`gevent.event.AsyncResult` object which may be
    used directly, though it will couple your code with gevent.
    """

    #: The encapsulated :class:`gevent.event.AsyncResult`
    async_result = None

    def __init__(self, async_result=None):
        super(GeventFuture, self).__init__()
        if async_result is not None:
            self.async_result = async_result
        else:
            self.async_result = gevent.event.AsyncResult()

    def get(self, timeout=None):
        try:
            return super(GeventFuture, self).get(timeout=timeout)
        except NotImplementedError:
            pass

        try:
            return self.async_result.get(timeout=timeout)
        except gevent.Timeout as e:
            raise Timeout(e)

    def set(self, value=None):
        assert not self.async_result.ready(), 'value has already been set'
        self.async_result.set(value)

    def set_exception(self, exc_info=None):
        if isinstance(exc_info, BaseException):
            exception = exc_info
        else:
            exc_info = exc_info or sys.exc_info()
            exception = exc_info[1]
        self.async_result.set_exception(exception)


class GeventActor(Actor):

    """
    :class:`GeventActor` implements :class:`pykka.Actor` using the `gevent
    <http://www.gevent.org/>`_ library. gevent is a coroutine-based Python
    networking library that uses greenlet to provide a high-level synchronous
    API on top of libevent event loop.

    This is a very fast implementation, but as of gevent 0.13.x it does not
    work in combination with other threads.
    """

    @staticmethod
    def _create_actor_inbox():
        return gevent.queue.Queue()

    @staticmethod
    def _create_future():
        return GeventFuture()

    def _start_actor_loop(self):
        gevent.Greenlet.spawn(self._actor_loop)



from __future__ import absolute_import

import sys
import threading

from pykka import compat
from pykka.actor import Actor
from pykka.exceptions import Timeout
from pykka.future import Future


__all__ = [
    'ThreadingActor',
    'ThreadingFuture',
]


class ThreadingFuture(Future):

    """
    :class:`ThreadingFuture` implements :class:`Future` for use with
    :class:`ThreadingActor <pykka.ThreadingActor>`.

    The future is implemented using a :class:`Queue.Queue`.

    The future does *not* make a copy of the object which is :meth:`set()
    <pykka.Future.set>` on it. It is the setters responsibility to only pass
    immutable objects or make a copy of the object before setting it on the
    future.

    .. versionchanged:: 0.14
        Previously, the encapsulated value was a copy made with
        :func:`copy.deepcopy`, unless the encapsulated value was a future, in
        which case the original future was encapsulated.
    """

    def __init__(self):
        super(ThreadingFuture, self).__init__()
        self._queue = compat.queue.Queue(maxsize=1)
        self._data = None

    def get(self, timeout=None):
        try:
            return super(ThreadingFuture, self).get(timeout=timeout)
        except NotImplementedError:
            pass

        try:
            if self._data is None:
                self._data = self._queue.get(True, timeout)
            if 'exc_info' in self._data:
                compat.reraise(*self._data['exc_info'])
            else:
                return self._data['value']
        except compat.queue.Empty:
            raise Timeout('%s seconds' % timeout)

    def set(self, value=None):
        self._queue.put({'value': value}, block=False)

    def set_exception(self, exc_info=None):
        if isinstance(exc_info, BaseException):
            exc_info = (exc_info.__class__, exc_info, None)
        self._queue.put({'exc_info': exc_info or sys.exc_info()})


class ThreadingActor(Actor):

    """
    :class:`ThreadingActor` implements :class:`Actor` using regular Python
    threads.

    This implementation is slower than :class:`GeventActor
    <pykka.gevent.GeventActor>`, but can be used in a process with other
    threads that are not Pykka actors.
    """

    use_daemon_thread = False
    """
    A boolean value indicating whether this actor is executed on a thread that
    is a daemon thread (:class:`True`) or not (:class:`False`). This must be
    set before :meth:`pykka.Actor.start` is called, otherwise
    :exc:`RuntimeError` is raised.

    The entire Python program exits when no alive non-daemon threads are left.
    This means that an actor running on a daemon thread may be interrupted at
    any time, and there is no guarantee that cleanup will be done or that
    :meth:`pykka.Actor.on_stop` will be called.

    Actors do not inherit the daemon flag from the actor that made it. It
    always has to be set explicitly for the actor to run on a daemonic thread.
    """

    @staticmethod
    def _create_actor_inbox():
        return compat.queue.Queue()

    @staticmethod
    def _create_future():
        return ThreadingFuture()

    def _start_actor_loop(self):
        thread = threading.Thread(target=self._actor_loop)
        thread.name = thread.name.replace('Thread', self.__class__.__name__)
        thread.daemon = self.use_daemon_thread
        thread.start()



#! /usr/bin/env python

import logging
import os
import signal
import time

import pykka
import pykka.debug


class DeadlockActorA(pykka.ThreadingActor):

    def foo(self, b):
        logging.debug('This is foo calling bar')
        return b.bar().get()


class DeadlockActorB(pykka.ThreadingActor):

    def __init__(self, a):
        super(DeadlockActorB, self).__init__()
        self.a = a

    def bar(self):
        logging.debug('This is bar calling foo; BOOM!')
        return self.a.foo().get()


if __name__ == '__main__':
    print('Setting up logging to get output from signal handler...')
    logging.basicConfig(level=logging.DEBUG)

    print('Registering signal handler...')
    signal.signal(signal.SIGUSR1, pykka.debug.log_thread_tracebacks)

    print('Starting actors...')
    a = DeadlockActorA.start().proxy()
    b = DeadlockActorB.start(a).proxy()

    print('Now doing something stupid that will deadlock the actors...')
    a.foo(b)

    time.sleep(0.01)  # Yield to actors, so we get output in a readable order

    print('Making main thread relax; not block, not quit')
    print('1) Use `kill -SIGUSR1 %d` to log thread tracebacks' % os.getpid())
    print('2) Then `kill %d` to terminate the process' % os.getpid())
    while True:
        time.sleep(1)



#! /usr/bin/env python

import pykka


class Adder(pykka.ThreadingActor):

    def add_one(self, i):
        print('{} is increasing {}'.format(self, i))
        return i + 1


class Bookkeeper(pykka.ThreadingActor):

    def __init__(self, adder):
        super(Bookkeeper, self).__init__()
        self.adder = adder

    def count_to(self, target):
        i = 0
        while i < target:
            i = self.adder.add_one(i).get()
            print('{} got {} back'.format(self, i))


if __name__ == '__main__':
    adder = Adder.start().proxy()
    bookkeeper = Bookkeeper.start(adder).proxy()
    bookkeeper.count_to(10).get()
    pykka.ActorRegistry.stop_all()



#! /usr/bin/env python

import pykka


class PlainActor(pykka.ThreadingActor):

    def __init__(self):
        super(PlainActor, self).__init__()
        self.stored_messages = []

    def on_receive(self, message):
        if message.get('command') == 'get_messages':
            return self.stored_messages
        else:
            self.stored_messages.append(message)


if __name__ == '__main__':
    actor = PlainActor.start()
    actor.tell({'no': 'Norway', 'se': 'Sweden'})
    actor.tell({'a': 3, 'b': 4, 'c': 5})
    print(actor.ask({'command': 'get_messages'}))
    actor.stop()
